/* libnfnetlink.c: Generic library for access to Netfilter via Netlink.
 *
 * Jay Schulist <jschlst@samba.org>, Copyright (c) 2001.
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <asm/types.h>
#include <linux/if.h>
#include <linux/netlink.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_tcp.h>
#include <linux/netfilter_ipv4/ip_conntrack_ftp.h>
#include <linux/netfilter_ipv4/ip_conntrack_icmp.h>
#include <linux/netfilter_ipv4/ip_conntrack_tuple.h>
#include <linux/nfnetlink.h>

#include "libnfnetlink.h"

#define nfnl_error(format, args...) \
	fprintf(stderr, format, ## args)

int nfnl_open(struct nfnl_handle *cth, unsigned subscriptions)
{
	int addr_len, err;

        memset(cth, 0, sizeof(cth));
        cth->fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_NETFILTER);
        if(cth->fd < 0) 
	{
                nfnl_error("socket(netlink): %s", strerror(errno));
                return (cth->fd);
        }

        memset(&cth->local, 0, sizeof(cth->local));
        cth->local.nl_family = AF_NETLINK;
        cth->local.nl_groups = subscriptions;

        err = bind(cth->fd, (struct sockaddr *)&cth->local, sizeof(cth->local));
	if(err < 0) 
	{
                nfnl_error("bind(netlink): %s", strerror(errno));
                return (err);
        }
        addr_len = sizeof(cth->local);
        err = getsockname(cth->fd, (struct sockaddr *)&cth->local, &addr_len);
	if(err < 0)
	{
		nfnl_error("getsockname failed(%d): %s\n", err,strerror(errno));
                return (err);
        }
        if(addr_len != sizeof(cth->local)) 
	{
                nfnl_error("Bad address length (%d)\n", addr_len);
                return (-1);
        }
        if(cth->local.nl_family != AF_NETLINK) 
	{
                nfnl_error("Bad address family (%d)\n", 
			cth->local.nl_family);
                return (-1);
        }
        cth->seq = time(NULL);

        return (0);
}

int nfnl_close(struct nfnl_handle *cth)
{
	if(cth->fd)
		close(cth->fd);
	memset(cth, 0, sizeof(cth));

	return (0);
}

int nfnl_send(struct nfnl_handle *cth, char *buf, int len)
{
        struct sockaddr_nl nladdr;

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;

        return (sendto(cth->fd, buf, len, 0, (struct sockaddr *)&nladdr, 
		sizeof(nladdr)));
}

int nfnl_listen(struct nfnl_handle *cth,
	int (*handler)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
        void *jarg)
{
	struct sockaddr_nl nladdr;
	char buf[NFNL_BUFFSIZE];
	struct nlmsghdr *h;
	struct iovec iov;
        int status;
        struct msghdr msg = {
                (void*)&nladdr, sizeof(nladdr),
                &iov, 1,
                NULL, 0,
                0
        };

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;
        nladdr.nl_pid	 = 0;
        nladdr.nl_groups = 0;
        iov.iov_base	 = buf;
        iov.iov_len  	 = sizeof(buf);

        while(1)
	{
                status = recvmsg(cth->fd, &msg, 0);
                if(status < 0)
		{
                        if(errno == EINTR)
                                continue;
			nfnl_error("nfnl_listen: recvmsg over-run\n");
                        continue;
                }
                if(status == 0)
		{
			nfnl_error("nfnl_listen: EOF on netlink\n");
                        return (-1);
                }
                if(msg.msg_namelen != sizeof(nladdr)) 
		{
			nfnl_error("nfnl_listen: Bad sender address len (%d)\n",
	 			msg.msg_namelen);
                        return (-1);
                }

                for(h = (struct nlmsghdr *)buf; status >= sizeof(*h);) 
		{
                        int err;
                        int len = h->nlmsg_len;
                        int l = len - sizeof(*h);

                        if(l < 0 || len > status) 
			{
                                if(msg.msg_flags & MSG_TRUNC) 
				{
					nfnl_error("nfnl_listen: MSG_TRUNC\n");
                                        return (-1);
                                }
				nfnl_error("nfnl_listen: Malformed msg (%d)\n",
					len);
                                return (-1);
                        }

                        err = handler(&nladdr, h, jarg);
                        if(err < 0)
                                return (err);

                        status -= NLMSG_ALIGN(len);
                        h = (struct nlmsghdr *)((char *)h + NLMSG_ALIGN(len));
                }
                if(msg.msg_flags & MSG_TRUNC) 
		{
			nfnl_error("nfnl_listen: MSG_TRUNC\n");
                        continue;
                }
                if(status) 
		{
			nfnl_error("nfnl_listen: Remnant size (%d)\n", status);
                        return (-1);
                }
        }

	return (0);
}

int nfnl_talk(struct nfnl_handle *cth, struct nlmsghdr *n, pid_t peer,
	unsigned groups, struct nlmsghdr *answer,
        int (*junk)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
        void *jarg)
{
	char buf[NFNL_BUFFSIZE];
	struct sockaddr_nl nladdr;
	struct nlmsghdr *h;
	unsigned seq;
        int status;
        struct iovec iov  = { 
		(void *)n, n->nlmsg_len 
	};
        struct msghdr msg = {
                (void*)&nladdr, sizeof(nladdr),
                &iov, 1,
                NULL, 0,
                0
        };

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;
        nladdr.nl_pid 	 = peer;
        nladdr.nl_groups = groups;

        n->nlmsg_seq = seq = ++cth->seq;
        if(answer == NULL)
                n->nlmsg_flags |= NLM_F_ACK;

        status = sendmsg(cth->fd, &msg, 0);
        if(status < 0) 
	{
		nfnl_error("nfnl_talk: sendmsg(netlink) %s\n",
			strerror(errno));
                return (-1);
        }

        iov.iov_base = buf;
        iov.iov_len = sizeof(buf);

        while(1) 
	{
                status = recvmsg(cth->fd, &msg, 0);
                if(status < 0) 
		{
                        if(errno == EINTR)
                                continue;
			nfnl_error("nfnl_talk: recvmsg over-run\n");
                        continue;
                }
                if(status == 0) 
		{
			nfnl_error("nfnl_talk: EOF on netlink\n");
                        return (-1);
                }
                if(msg.msg_namelen != sizeof(nladdr)) 
		{
			nfnl_error("nfnl_talk: Bad sender address len (%d)\n",
                                msg.msg_namelen);
                        return (-1);
                }

                for(h = (struct nlmsghdr *)buf; status >= sizeof(*h);) 
		{
                        int len = h->nlmsg_len;
                        int l = len - sizeof(*h);
			int err;

                        if(l < 0 || len > status) 
			{
                                if(msg.msg_flags & MSG_TRUNC) 
				{
					nfnl_error("nfnl_talk: MSG_TRUNC\n");
                                        return (-1);
                                }
				nfnl_error("nfnl_talk: Malformed msg (%d)\n", 
					len);
                                return (-1);
                        }

                        if(h->nlmsg_pid != cth->local.nl_pid
                            || h->nlmsg_seq != seq) 
			{
                                if(junk) 
				{
                                        err = junk(&nladdr, h, jarg);
                                        if(err < 0)
                                                return (err);
                                }
                                continue;
                        }

                        if(h->nlmsg_type == NLMSG_ERROR) 
			{
                                struct nlmsgerr *nerr 
					= (struct nlmsgerr *)NLMSG_DATA(h);
                                if(l < sizeof(struct nlmsgerr))
					nfnl_error("nfnl_talk: ERROR MSG_TRUNC\n");
                                else 
				{
                                        errno = -nerr->error;
                                        if(errno == 0) 
					{
                                                if(answer)
                                                        memcpy(answer, h, h->nlmsg_len);
                                                return (0);
                                        }
                                }
                                return (-1);
                        }
                        if(answer)
			{
                                memcpy(answer, h, h->nlmsg_len);
                                return (0);
                        }

			nfnl_error("nfnl_talk: Unexpected reply\n");

                        status -= NLMSG_ALIGN(len);
                        h = (struct nlmsghdr *)((char *)h + NLMSG_ALIGN(len));
                }
                if(msg.msg_flags & MSG_TRUNC)
		{
			nfnl_error("nfnl_talk: MSG_TRUNC\n");
                        continue;
                }
                if(status) 
		{
			nfnl_error("nfnl_talk: Remnant size (%d)\n", status);
                        return (-1);
                }
        }

	return (0);
}

int nfnl_wilddump_request(struct nfnl_handle *cth, int family, int type)
{
        struct {
                struct nlmsghdr nlh;
                struct nfgenmsg g;
        } req;
        struct sockaddr_nl nladdr;

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;

        req.nlh.nlmsg_len	= sizeof(req);
        req.nlh.nlmsg_type 	= type;
        req.nlh.nlmsg_flags 	= NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
        req.nlh.nlmsg_pid 	= 0;
        req.nlh.nlmsg_seq 	= cth->dump = ++cth->seq;
        req.g.nfgen_family 	= family;

        return (sendto(cth->fd, (void *)&req, sizeof(req), 0, 
		(struct sockaddr *)&nladdr, sizeof(nladdr)));
}

int nfnl_dump_request(struct nfnl_handle *cth, int type, void *req, int len)
{
	struct sockaddr_nl nladdr;
        struct nlmsghdr nlh;
        struct iovec iov[2] = { 
		{ &nlh, sizeof(nlh) }, 
		{ req, len } 
	};
        struct msghdr msg = {
                (void*)&nladdr, sizeof(nladdr),
                iov,  2,
                NULL, 0,
                0
        };

        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;

        nlh.nlmsg_len	= NLMSG_LENGTH(len);
        nlh.nlmsg_type 	= type;
        nlh.nlmsg_flags = NLM_F_ROOT|NLM_F_MATCH|NLM_F_REQUEST;
        nlh.nlmsg_pid 	= 0;
        nlh.nlmsg_seq =	 cth->dump = ++cth->seq;

        return (sendmsg(cth->fd, &msg, 0));
}

int nfnl_dump_filter(struct nfnl_handle *cth,
	int (*filter)(struct sockaddr_nl *, struct nlmsghdr *n, void *),
        void *arg1,
        int (*junk)(struct sockaddr_nl *,struct nlmsghdr *n, void *),
        void *arg2)
{
	struct sockaddr_nl nladdr;
	char buf[NFNL_BUFFSIZE];
        struct iovec iov = { 
		buf, sizeof(buf) 
	};

        while(1) 
	{
		struct nlmsghdr *h;
                int status;
                struct msghdr msg = {
                        (void*)&nladdr, sizeof(nladdr),
                        &iov, 1,
                        NULL, 0,
                        0
                };

                status = recvmsg(cth->fd, &msg, 0);
                if(status < 0)
		{
                        if(errno == EINTR)
                                continue;
			nfnl_error("nfnl_dump_filter: recvmsg over-run\n");
                        continue;
                }
                if(status == 0) 
		{
			nfnl_error("nfnl_dump_filter: EOF on netlink\n");
                        return (-1);
                }
                if(msg.msg_namelen != sizeof(nladdr)) 
		{
			nfnl_error("nfnl_dump_filter: Bad sender address len (%d)\n",
                                msg.msg_namelen);
                        return (-1);
                }

                h = (struct nlmsghdr *)buf;
                while(NLMSG_OK(h, status)) 
		{
                        int err;

                        if(h->nlmsg_pid != cth->local.nl_pid
                        	|| h->nlmsg_seq != cth->dump) 
			{
                                if(junk) 
				{
                                        err = junk(&nladdr, h, arg2);
                                        if(err < 0)
                                                return (err);
                                }
                                goto skip_it;
                        }

                        if(h->nlmsg_type == NLMSG_DONE)
                                return (0);
                        if(h->nlmsg_type == NLMSG_ERROR) 
			{
                                struct nlmsgerr *nerr 
					= (struct nlmsgerr *)NLMSG_DATA(h);
                                if(h->nlmsg_len < NLMSG_LENGTH(sizeof(struct nlmsgerr)))
					nfnl_error("nfnl_dump_filter: MSG_TRUNC\n");
                                else 
                                        errno = -nerr->error;
                                return (-1);
                        }

                        err = filter(&nladdr, h, arg1);
                        if(err < 0)
                                return (err);

skip_it:
                        h = NLMSG_NEXT(h, status);
                }

                if(msg.msg_flags & MSG_TRUNC)
		{
			nfnl_error("nfnl_dump_filter: MSG_TRUNC\n");
                        continue;
                }
                if(status) 
		{
			nfnl_error("nfnl_dump_filter: Remnant size (%d)\n", 
				status);
                        return (-1);
                }
        }

	return (0);
}

int nfnl_addattr32(struct nlmsghdr *n, int maxlen, int type, u_int32_t data)
{
        int len = NFA_LENGTH(4);
        struct nfattr *cta;

        if((NLMSG_ALIGN(n->nlmsg_len) + len) > maxlen)
                return (-1);
        cta = (struct nfattr *)(((char *)n) + NLMSG_ALIGN(n->nlmsg_len));
        cta->nfa_type = type;
        cta->nfa_len  = len;
        memcpy(NFA_DATA(cta), &data, 4);
        n->nlmsg_len = (NLMSG_ALIGN(n->nlmsg_len) + len);

        return (0);
}

int nfnl_addattr_l(struct nlmsghdr *n, int maxlen, int type, void *data, 
	int alen)
{
        int len = NFA_LENGTH(alen);
        struct nfattr *cta;

        if((NLMSG_ALIGN(n->nlmsg_len) + len) > maxlen)
                return (-1);
        cta = (struct nfattr *)(((char *)n) + NLMSG_ALIGN(n->nlmsg_len));
        cta->nfa_type = type;
        cta->nfa_len  = len;
        memcpy(NFA_DATA(cta), data, alen);
        n->nlmsg_len = (NLMSG_ALIGN(n->nlmsg_len) + len);

        return (0);
}

int nfnl_cta_addattr32(struct nfattr *cta, int maxlen, int type, u_int32_t data)
{
        int len = NFA_LENGTH(4);
        struct nfattr *subcta;

        if((NFA_ALIGN(cta->nfa_len) + len) > maxlen)
                return (-1);
        subcta = (struct nfattr *)(((char *)cta) + NFA_ALIGN(cta->nfa_len));
        subcta->nfa_type = type;
        subcta->nfa_len  = len;
        memcpy(NFA_DATA(subcta), &data, 4);
        cta->nfa_len = (NLMSG_ALIGN(cta->nfa_len) + len);

        return (0);
}

int nfnl_cta_addattr_l(struct nfattr *cta, int maxlen, int type, void *data, 
	int alen)
{
        struct nfattr *subcta;
        int len = NFA_LENGTH(alen);

        if((NFA_ALIGN(cta->nfa_len) + len) > maxlen)
                return (-1);
        subcta = (struct nfattr *)(((char *)cta) + NFA_ALIGN(cta->nfa_len));
        subcta->nfa_type = type;
        subcta->nfa_len  = len;
        memcpy(NFA_DATA(subcta), data, alen);
        cta->nfa_len = (NLMSG_ALIGN(cta->nfa_len) + len);

        return (0);
}

int nfnl_parse_attr(struct nfattr *tb[], int max, struct nfattr *cta, int len)
{
        while(NFA_OK(cta, len))
	{
                if(cta->nfa_type <= max)
                        tb[cta->nfa_type] = cta;
                cta = NFA_NEXT(cta,len);
        }
        if(len)
		nfnl_error("nfnl_parse_attr: deficit (%d) len (%d).\n",
			len, cta->nfa_len);
        return (0);
}
