/* nfconntrack.c: User interface to the connection tracking netlink interface.
 *
 * Jay Schulist <jschlst@samba.org>, Copyright (c) 2001.
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 */

#include <version.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <asm/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netinet/ip_icmp.h>

#include <linux/if.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <libnfnetlink.h>

#include "nf.h"
#include "utility.h"
#include "display.h"

extern void nfconntrack_usage(void);

static struct
{
        struct nfnl_handle *cth;
        int protocol;
	int family;
	char in_dev[IFNAMSIZ];
	char out_dev[IFNAMSIZ];
	int ip_where;
	struct in_addr addr;
        struct in_addr mask;
} filter;

#define IPW_ORIG	0x1
#define IPW_RPLY	0x2
#define IPW_SRC		0x4
#define IPW_DST		0x8

void ctrace_reset_filter(void)
{
        memset(&filter, 0, sizeof(filter));
	filter.in_dev[0]  = -1;
	filter.out_dev[0] = -1;
	filter.ip_where   = -1;
}

void incomplete_command()
{
        fprintf(stderr, "Command line is not complete. Try option \"help\"\n");
        exit(-1);
}

#define NEXT_ARG() do { argv++; if (--argc <= 0) incomplete_command(); } while(0)

int ctrace_run_filter(struct ctmsg *c, struct nfattr *cb[])
{
	struct cta_proto *cp;

	if(filter.protocol)
        {
                if(!cb[CTA_PROTOINFO])
                        return (1);
                cp = NFA_DATA(cb[CTA_PROTOINFO]);
                if(cp->num_proto != filter.protocol)
                        return (1);
        }
        if(filter.family)
        {
                if(c->ctm_family != filter.family)
                        return (1);
        }
        if(filter.in_dev[0] > 0)
        {
                if(!cb[CTA_IIF])
                        return (1);
                if(strncmp(filter.in_dev, (char *)NFA_DATA(cb[CTA_IIF]),
                        IFNAMSIZ))
                        return (1);
        }
	if(filter.out_dev[0] > 0)
        {
                if(!cb[CTA_OIF])
                        return (1);
                if(strncmp(filter.out_dev, (char *)NFA_DATA(cb[CTA_OIF]),
                        IFNAMSIZ))
                        return (1);
        }
        if(filter.ip_where > 0)
        {
                struct ip_conntrack_tuple *tuple;

                if(filter.ip_where & IPW_ORIG)
                {
                        if(!cb[CTA_ORIG])
                                return (1);
                        tuple = NFA_DATA(cb[CTA_ORIG]);
                }
                else
                {
                        if(filter.ip_where & IPW_RPLY)
                        {
                                if(!cb[CTA_RPLY])
                                        return (1);
                                tuple = NFA_DATA(cb[CTA_RPLY]);
                        }
                }

                if(filter.ip_where & IPW_DST)
                {
                        if((tuple->dst.ip & filter.mask.s_addr)
                                != (filter.addr.s_addr & filter.mask.s_addr))
                                return (1);
                }
                else
                {
                        if(filter.ip_where & IPW_SRC)
                        {
                                if((tuple->src.ip & filter.mask.s_addr)
                                        != (filter.addr.s_addr
                                        & filter.mask.s_addr))
                                        return (1);
                        }
                }
        }

	return (0);	/* Okay packet. */
}

int ctrace_flush_conntrack(struct sockaddr_nl *who,struct nlmsghdr *n,void *arg)
{
	struct nfattr *cb[CTA_MAX + 1];
        struct ctmsg *c = NLMSG_DATA(n);
        int len = n->nlmsg_len;
        struct nfnl_handle cth;

        struct {
                struct nlmsghdr nh;
                struct ctmsg ct;
                char buf[1024];
        } cdr;

	if(len > sizeof(cdr))
		return (0);
	memcpy(&cdr.nh, n, n->nlmsg_len);

	/* check data and copy here. */
        len -= NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct ctmsg)));
        if(len < 0)
        {
                printf("BUG: wrong nlmsg len (%d).", len);
                return (len);
        }

        memset(cb, 0, sizeof(cb));
        nfnl_parse_attr(cb, CTA_MAX, NFM_NFA(c), len);
	if(ctrace_run_filter(c, cb))
		return (0);

	if(nfnl_open(&cth, 0) < 0)
                exit (1);
	cdr.nh.nlmsg_type = NFM_DELCONNTRACK;
	cdr.nh.nlmsg_flags = NLM_F_REQUEST;
        if(nfnl_talk(&cth, &cdr.nh, 0, 0, NULL, NULL, NULL) < 0)
                exit(2);
        nfnl_close(&cth);

	return (0);
}

/* display an entire ctnetlink message to the user. */
int ctrace_print_conntrack(struct sockaddr_nl *who,struct nlmsghdr *n,void *arg)
{
	struct nfattr *cb[CTA_MAX + 1];
	struct ctmsg *c = NLMSG_DATA(n);
        int len = n->nlmsg_len;

	/* check data and copy here. */
        len -= NLMSG_LENGTH(NLMSG_ALIGN(sizeof(struct ctmsg)));
        if(len < 0)
        {
       		printf("BUG: wrong nlmsg len (%d).", len);
                return (len);
        }

        memset(cb, 0, sizeof(cb));
	nfnl_parse_attr(cb, CTA_MAX, NFM_NFA(c), len);

	if(ctrace_run_filter(c, cb))
		return (0);

	if(cb[CTA_ORIG] && cb[CTA_RPLY])
	{
		printf("%s, ", 
			display_tuple_flat(NFA_DATA(cb[CTA_ORIG])));
		printf("%s",
			display_tuple_flat(NFA_DATA(cb[CTA_RPLY])));
	}
	else
	{
		if(cb[CTA_ORIG])
			printf("%s (Orig)", 
				display_tuple_flat(NFA_DATA(cb[CTA_ORIG])));
		if(cb[CTA_RPLY])
			printf("%s (Rply)", 
				display_tuple_flat(NFA_DATA(cb[CTA_RPLY]))); 
	}

	printf("\n");

	if(cb[CTA_PROTOINFO])
	{
		struct cta_proto *cp = NFA_DATA(cb[CTA_PROTOINFO]);

		printf("%s %s", pr_af_family(c->ctm_family), 
			pr_proto(cp->num_proto));

		if(cp->num_proto == IPPROTO_TCP)
	                printf(" (%s)", 
				pr_tcp_state(cp->proto.tcp.state));
	}

	if(cb[CTA_IIF])
		printf(", In %s", (char *)NFA_DATA(cb[CTA_IIF]));

	if(cb[CTA_OIF])
		printf(", Out %s", (char *)NFA_DATA(cb[CTA_OIF]));

	if(cb[CTA_INFO])
	{
		unsigned long *info = NFA_DATA(cb[CTA_INFO]);
		printf(", %s", pr_ctinfo(*info));
	}

        if(cb[CTA_STATUS])
        {
                unsigned long *status = NFA_DATA(cb[CTA_STATUS]);
                printf(", %s", pr_ct_status(*status));
        }

	printf("\n");

	if(cb[CTA_NATINFO])
	{
		struct cta_nat *cn = NFA_DATA(cb[CTA_NATINFO]);
		int i;

		printf("NAT (%d): ", cn->num_manips);
		for(i = 0; i < cn->num_manips; i++)
		{
			struct ip_nat_info_manip *m
				= &cn->manips[i];
			printf("%s %s: %s:%d", 
				pr_man_direction(m->direction),
				pr_man_type(m->maniptype),
				in_ntoa(m->manip.ip), 
				ntohs(m->manip.u.all));
			if((i + 1) < cn->num_manips)
				printf(", ");
		}
		printf("\n");
	}


	if(cb[CTA_HELPINFO])
	{
		printf("Helper information present.\n");
	}

        return (0);
}

static int ctrace_get_tuple(int argc, char **argv, 
	struct ip_conntrack_tuple *tuple)
{
	if(argc < 5)
	{
		fprintf(stderr, "Must specify at least TUPLE\n");
                exit (1);
	}

	if(matches(*argv, "udp") == 0)
		tuple->dst.protonum = IPPROTO_UDP;
        else {
        if(matches(*argv, "tcp") == 0)
		tuple->dst.protonum = IPPROTO_TCP;
        else
        if(matches(*argv, "icmp") == 0)
		tuple->dst.protonum = IPPROTO_ICMP;
        }

	NEXT_ARG();
        tuple->src.ip           = in_aton(*argv);
	NEXT_ARG();
        tuple->src.u.all        = htons(atoi(*argv));
	NEXT_ARG();
        tuple->dst.ip           = in_aton(*argv);
	NEXT_ARG();
        tuple->dst.u.all        = htons(atoi(*argv));

        return (0);
}

#define INC_ARG()	{ (void)*argv++; argc--; }
static int ctrace_load_filter(int argc, char **argv)
{
	ctrace_reset_filter();

	while(argc > 0)
        {
        	if(matches(*argv, "udp") == 0)
		{
	                filter.protocol = IPPROTO_UDP;
			INC_ARG();
			continue;
		}

                if(matches(*argv, "tcp") == 0)
                {
			filter.protocol = IPPROTO_TCP;
			INC_ARG();
                        continue;
		}

                if(matches(*argv, "icmp") == 0)
                {
			filter.protocol = IPPROTO_ICMP;
			INC_ARG();
                        continue;
                }

		if(matches(*argv, "ip") == 0)
                {
			char *ip, *mask;
                        if(*++argv == NULL) nfconntrack_usage();
                        argc--;

			if(matches(*argv, "orig") == 0)
				filter.ip_where = IPW_ORIG;
			else 
			{
				if(matches(*argv, "rply") == 0)
	                                filter.ip_where = IPW_RPLY;
				else
					nfconntrack_usage();
			}
			NEXT_ARG();

			if(matches(*argv, "dst") == 0)
				filter.ip_where |= IPW_DST;
			else
			{
				if(matches(*argv, "src") == 0)
					filter.ip_where |= IPW_SRC;
				else
                                        nfconntrack_usage();
			}
			NEXT_ARG();

			ip = strtok(*argv, "/");
                        mask = strtok(NULL, "/");
                        if(ip == NULL || mask == NULL)
                                break;

			inet_aton(ip, &filter.addr);
			switch(atoi(mask)) {
                                case (8):
                                        filter.mask.s_addr = 0x000000ff;
                                        break;
                                case (16):
                                        filter.mask.s_addr = 0x0000ffff;
                                        break;
                                case (24):
                                        filter.mask.s_addr = 0x00ffffff;
                                        break;
                                case (32):
                                        filter.mask.s_addr = 0xffffffff;
                                        break;
                                default:
                                        break;
                        }

                        INC_ARG();
                        continue;
                }

		if(matches(*argv, "ipv4") == 0)
		{
			filter.family = AF_INET;
			INC_ARG();
                        continue;
                }

		if(matches(*argv, "ipv6") == 0)
                {
			filter.family = AF_INET6;
                        INC_ARG();
                        continue;
                }

		if(matches(*argv, "in") == 0)
                {
			if(*++argv == NULL) nfconntrack_usage();
			argc--;
			strncpy(filter.in_dev, *argv, IFNAMSIZ);
                        INC_ARG();
                        continue;
                }

		if(matches(*argv, "out") == 0)
                {
			if(*++argv == NULL) nfconntrack_usage();
			argc--;
			strncpy(filter.out_dev, *argv, IFNAMSIZ);
                        INC_ARG();
                        continue;
                }

                argv++;
                argc--;
        }

	return (0);
}

static int ctrace_list_or_flush(int argc, char **argv, int flush)
{
	struct nfnl_handle cth;

	ctrace_load_filter(argc, argv);
	if(nfnl_open(&cth, 0) < 0)
		exit (1);

	if(flush)
	{
		if(nfnl_wilddump_request(&cth, AF_INET, NFM_GETCONNTRACK) < 0)
			exit (1);
		if(nfnl_dump_filter(&cth, ctrace_flush_conntrack, NULL, NULL, 
			NULL) < 0)
			exit (1);
		nfnl_close(&cth);
		exit (0);
	}

	if(nfnl_wilddump_request(&cth, AF_INET, NFM_GETCONNTRACK) < 0)
		exit (1);
	if(nfnl_dump_filter(&cth, ctrace_print_conntrack, stdout, NULL, NULL) 
		< 0)
		exit (1);

	nfnl_close(&cth);
	exit (0);	
}

int ctrace_modify(int cmd, unsigned flags, int argc, char **argv)
{
	struct ip_conntrack_tuple tuple;
	struct sockaddr_nl to;
	struct nfnl_handle cth;

        struct {
                struct nlmsghdr nh;
                struct ctmsg ct;
                char buf[1024];
        } cdr;

        memset(&to, 0, sizeof(to));
        to.nl_family = AF_NETLINK;
        cdr.nh.nlmsg_len      = NLMSG_LENGTH(sizeof(struct ctmsg));
        cdr.nh.nlmsg_type     = cmd;
        cdr.nh.nlmsg_flags    = NLM_F_REQUEST|flags;
        cdr.nh.nlmsg_pid      = 0;
        cdr.nh.nlmsg_seq      = 0;
        cdr.ct.ctm_family     = AF_INET;
        cdr.ct.ctm_orig_len   = sizeof(struct ip_conntrack_tuple);
	ctrace_get_tuple(argc, argv, &tuple);
        nfnl_addattr_l(&cdr.nh, sizeof(cdr), CTA_ORIG, &tuple,
                sizeof(struct ip_conntrack_tuple));

	if(nfnl_open(&cth, 0) < 0)
		exit (1);
	if(nfnl_talk(&cth, &cdr.nh, 0, 0, &cdr.nh, NULL, NULL) < 0)
                exit(2);

	nfnl_close(&cth);
	exit (0);
}

int ctrace_get(int argc, char **argv)
{
	struct ip_conntrack_tuple tuple;
	struct nfnl_handle cth;
	struct sockaddr_nl to;

        struct {
                struct nlmsghdr nh;
                struct ctmsg ct;
                char buf[1024];
        } cdr;

        memset(&to, 0, sizeof(to));
        to.nl_family = AF_NETLINK;
        cdr.nh.nlmsg_len      = NLMSG_LENGTH(sizeof(struct ctmsg));
        cdr.nh.nlmsg_type     = NFM_GETCONNTRACK;
        cdr.nh.nlmsg_flags    = NLM_F_REQUEST;
        cdr.nh.nlmsg_pid      = 0;
        cdr.nh.nlmsg_seq      = 0;
        cdr.ct.ctm_family     = AF_INET;
        cdr.ct.ctm_orig_len   = sizeof(struct ip_conntrack_tuple);
	ctrace_get_tuple(argc, argv, &tuple);
	nfnl_addattr_l(&cdr.nh, sizeof(cdr), CTA_ORIG, &tuple,
		sizeof(struct ip_conntrack_tuple));

	if(nfnl_open(&cth, 0) < 0)
		exit (1);
	if(nfnl_talk(&cth, &cdr.nh, 0, 0, &cdr.nh, NULL, NULL) < 0)
                exit(2);

	ctrace_print_conntrack(NULL, &cdr.nh, (void*)stdout);
	nfnl_close(&cth);
	exit (0);
}

int ctrace_monitor(int argc, char **argv)
{
	struct nfnl_handle cth;
	int groups = 0;

	if(argc < 2)
	{
		fprintf(stderr, "Must specify at least one proto CTPROTO\n");
		exit (1);
	}

	while(argc > 0)
	{
		if(matches(*argv, "proto") == 0)
		{
			NEXT_ARG();
			if(matches(*argv, "udp") == 0)
				groups |= NFGRP_IPV4_CT_UDP;
			else {
			if(matches(*argv, "tcp") == 0)
				groups |= NFGRP_IPV4_CT_TCP;
			else
			if(matches(*argv, "icmp") == 0)
				groups |= NFGRP_IPV4_CT_ICMP;
			}
		}

		if(matches(*argv, "match") == 0)
		{
			NEXT_ARG();
			ctrace_load_filter(argc, argv);
		}

		argv++;
		argc--;
	}

        if(nfnl_open(&cth, groups) < 0)
                exit (1);

	nfnl_listen(&cth, ctrace_print_conntrack, stdout);
	
	nfnl_close(&cth);
        exit (0);
}

/* display useless help information. */
void nfconntrack_usage(void)
{
        printf("Usage: nf conntrack { list | show | flush } SELECTOR\n");
        printf("       nf conntrack monitor proto CTPROTO SELECTOR\n");
        printf("       nf conntrack get TUPLE\n");
        printf("       nf conntrack { add | delete } TUPLE\n");
        printf("       nf conntrack { help }\n");
        printf("SELECTOR := [ match PREFIX ] [ exact PREFIX ]\n");
        printf("            [ proto CTPROTO ]\n");
        printf("TUPLE := [ CTPROTO SRC_IP SRC_PORT DST_IP DST_PORT ]\n");
        printf("PREFIX := [ CTPROTO &| FAMILY &| IPADDR &| DEVICE ]\n");
        printf("CTPROTO := [ udp | tcp | icmp ]\n");
        printf("FAMILY := [ ipv4 | ipv6 ]\n");
        printf("IPADDR := [ ip (orig|rply) (src|dst) xxx.xxx.xxx.xxx/X ]\n");
        printf("DEVICE := [ in NAME | out NAME ]\n");
        exit(1);
}

int nfconntrack(int family, int argc, char **argv)
{
	if(argc < 1)
		return (ctrace_list_or_flush(0, NULL, 0));
	if(matches(*argv, "add") == 0)
		return (ctrace_modify(NFM_NEWCONNTRACK, NLM_F_CREATE,
			argc - 1, argv + 1));
	if(matches(*argv, "delete") == 0
		|| matches(*argv, "del") == 0)
		return (ctrace_modify(NFM_DELCONNTRACK, NLM_F_ACK, argc - 1, 
		argv + 1));
	if(matches(*argv, "get") == 0)
		return (ctrace_get(argc - 1, argv + 1));
	if(matches(*argv, "monitor") == 0)
		return (ctrace_monitor(argc - 1, argv + 1));
	if(matches(*argv, "list") == 0 || matches(*argv, "show") == 0
		|| matches(*argv, "lst") == 0)
                return (ctrace_list_or_flush(argc - 1, argv + 1, 0));
	if(matches(*argv, "flush") == 0)
                return (ctrace_list_or_flush(argc - 1, argv + 1, 1));
	if(matches(*argv, "help") == 0)
                nfconntrack_usage();
        fprintf(stderr, "Command \"%s\" is unknown, try \"nf conntrack help\".\n", 
		*argv);

        exit (-1);
}
